# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A module providing process output sequence assertions.

To prevent pytest from rewriting this module assertions, please PYTEST_DONT_REWRITE.
See https://docs.pytest.org/en/latest/assert.html#disabling-assert-rewriting for
further reference.
"""


from contextlib import contextmanager

from ..util import resolveProcesses


class SequentialTextChecker:
    """Helper class for asserting that text is found in a certain order."""

    def __init__(self, output):
        self._output = output
        self._array_index = 0  # Keeps track of how far we are into the output array
        self._substring_index = 0  # Keeps track of how far we are into an individual string

    def assertInText(self, msg):
        return self.assertInStdout(msg)

    def assertInStdout(self, msg):

        # Keep local copies of the array index and the substring index.  We only advance them
        # if we find a matching string.

        array_index = self._array_index
        substring_index = self._substring_index

        for text in self._output[array_index:]:
            found = text.find(msg, substring_index)

            if found != -1:
                # We found the string!  Update the search state for the next string
                substring_index = found + len(msg)
                self._array_index = array_index
                self._substring_index = substring_index
                return

            # We failed to find the string.  Go around the loop again
            array_index += 1
            substring_index = 0

        assert False, (
            "'{}' not found in sequence after previous match.  "
            'The output near the last matched line:\n{}'
        ).format(
            msg,
            self.get_nearby_lines()
        )

    def get_nearby_lines(self):

        # This works by concatinating a few of the process_io outputs that we received, then
        # searching forward and backward for two return-lines in each direction, then returning
        # just that portion to give context about where a failure happened

        start_idx = max(0, self._array_index - 2)
        end_idx = self._array_index + 4

        full_text = ''.join(self._output[start_idx:end_idx])

        current_absolute_position = self._substring_index
        for txt in self._output[start_idx:self._array_index]:
            current_absolute_position += len(txt)

        start_abs_position = current_absolute_position
        for _ in range(3):
            start_abs_position = max(full_text.rfind('\n', 0, start_abs_position), 0)

        if start_abs_position != 0:
            # Complicated, but if the start_abs_position is not zero, it's pointed to a \n
            # we need to increment by 1 so the leading \n is not in the resulting text
            start_abs_position += 1

        end_abs_position = current_absolute_position
        for _ in range(3):
            end_abs_position = max(full_text.find('\n', end_abs_position + 1), end_abs_position)

        return full_text[start_abs_position:end_abs_position]


@contextmanager
def assertSequentialStdout(proc_output,
                           process,
                           cmd_args=None):
    """
    Create a context manager used to check stdout occured in a specific order.

    :param proc_output:  The captured output from a test run

    :param process: The process whose output will be searched
    :type process: A string (search by process name) or a launch.actions.ExecuteProcess object

    :param cmd_args: Optional.  If 'proc' is a string, cmd_args will be used to disambiguate
    processes with the same name.  Pass launch_testing.asserts.NO_CMD_ARGS to match a proc without
    command arguments
    :type cmd_args: string
    """
    process = resolveProcesses(
        proc_output,
        process=process,
        cmd_args=cmd_args,
        # There's no good way to sequence output from multiple processes reliably, so we won't
        # pretend to be able to.  Only allow one matching process for the comination of proc and
        # cmd_args
        strict_proc_matching=True,
    )[0]

    # Get all the output from the process.  This will be a list of strings.  Each string may
    # contain multiple lines of output
    to_check = [p.text.decode() for p in proc_output[process]]
    checker = SequentialTextChecker(to_check)

    try:
        yield checker
    except Exception:
        # Do we need to log this, or give a better message?  Need to re-raise so we can
        # cause the test to fail
        raise
    finally:
        # We don't need to do anything to finalize the checker here
        pass



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A module providing process output assertions.

To prevent pytest from rewriting this module assertions, please PYTEST_DONT_REWRITE.
See https://docs.pytest.org/en/latest/assert.html#disabling-assert-rewriting for
further reference.
"""

from osrf_pycommon.terminal_color import remove_ansi_escape_sequences

from ..tools.text import build_text_match
from ..util import resolveProcesses


def assertInStream(proc_output,
                   expected_output,
                   process,
                   cmd_args=None,
                   *,
                   output_filter=None,
                   strict_proc_matching=True,
                   strip_ansi_escape_sequences=True,
                   stream='stderr'):
    """
    Assert that 'output' was found in a stream of a process.

    :param proc_output: The process output captured by launch_test.  This is usually injected
    into test cases as self._proc_output
    :type proc_output: An launch_testing.IoHandler

    :param expected_output: The output to search for
    :type expected_output: string or regex pattern or a list of the aforementioned types

    :param process: The process whose output will be searched
    :type process: A string (search by process name) or a launch.actions.ExecuteProcess object

    :param cmd_args: Optional.  If 'process' is a string, cmd_args will be used to disambiguate
    processes with the same name.  Pass launch_testing.asserts.NO_CMD_ARGS to match a proc without
    command arguments
    :type cmd_args: string

    :param output_filter: Optional. A function to filter output before attempting any assertion.
    :type output_filter: callable

    :param strict_proc_matching: Optional (default True), If proc is a string and the combination
    of proc and cmd_args matches multiple processes, then strict_proc_matching=True will raise
    an error.
    :type strict_proc_matching: bool

    :param strip_ansi_escape_sequences: If True (default), strip ansi escape
    sequences from actual output before comparing with the output filter or
    expected output.
    :type strip_ansi_escape_sequences: bool

    :param stream: Which stream to examine.  This must be one of 'stderr' or 'stdout'.
    :type stream: string
    """
    resolved_procs = resolveProcesses(
        info_obj=proc_output,
        process=process,
        cmd_args=cmd_args,
        strict_proc_matching=strict_proc_matching
    )
    if output_filter is not None:
        if not callable(output_filter):
            raise ValueError('output_filter is not callable')
    output_match = build_text_match(expected_output)

    for proc in resolved_procs:  # Nominally just one matching proc
        if stream == 'stdout':
            full_output = ''.join(
                output.text.decode() for output in proc_output[proc] if output.from_stdout
            )
        elif stream == 'stderr':
            full_output = ''.join(
                output.text.decode() for output in proc_output[proc] if output.from_stderr
            )
        else:
            raise ValueError("Invalid value for stream; must be 'stdout' or 'stderr'")

        if strip_ansi_escape_sequences:
            full_output = remove_ansi_escape_sequences(full_output)
        if output_filter is not None:
            full_output = output_filter(full_output)
        if output_match(full_output) is not None:
            break
    else:
        names = ', '.join(sorted(p.process_details['name'] for p in resolved_procs))
        assert False, "Did not find '{}' in output for any of the matching processes: {}".format(
            expected_output, names
        )


def assertInStdout(proc_output,
                   expected_output,
                   process,
                   cmd_args=None,
                   *,
                   output_filter=None,
                   strict_proc_matching=True,
                   strip_ansi_escape_sequences=True):
    """
    Assert that 'output' was found in the standard output of a process.

    See the documentation for 'assertInStream' for full details.
    """
    assertInStream(proc_output, expected_output, process, cmd_args, output_filter=output_filter,
                   strict_proc_matching=strict_proc_matching,
                   strip_ansi_escape_sequences=strip_ansi_escape_sequences, stream='stdout')


def assertInStderr(proc_output,
                   expected_output,
                   process,
                   cmd_args=None,
                   *,
                   output_filter=None,
                   strict_proc_matching=True,
                   strip_ansi_escape_sequences=True):
    """
    Assert that 'output' was found in the standard error of a process.

    See the documentation for 'assertInStream' for full details.
    """
    assertInStream(proc_output, expected_output, process, cmd_args, output_filter=output_filter,
                   strict_proc_matching=strict_proc_matching,
                   strip_ansi_escape_sequences=strip_ansi_escape_sequences, stream='stderr')


def assertDefaultStream():
    """
    Return the stream that is used by default for 'assertInStream'.

    This is useful for writing tests that are compatible with both Eloquent and newer releases.
    """
    return 'stderr'



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import unittest

import ament_index_python

import launch
import launch.actions
import launch_testing
import launch_testing.actions
import launch_testing.util

import pytest


@pytest.mark.launch_test
@launch_testing.parametrize('arg_param', ['thing=On', 'thing=Off', 'flag1'])
def generate_test_description(arg_param):

    terminating_process = launch.actions.ExecuteProcess(
        cmd=[
            sys.executable,
            os.path.join(
                ament_index_python.get_package_prefix('launch_testing'),
                'lib/launch_testing',
                'terminating_proc.py',
            ),
            # Use the parameter passed to generate_test_description as an argument
            # to the terminating_proc
            '--{}'.format(arg_param),
        ]
    )

    return (
        launch.LaunchDescription([
            terminating_process,
            launch_testing.util.KeepAliveProc(),
            launch_testing.actions.ReadyToTest(),
        ]),
        {'dut_process': terminating_process}
    )


class TestProcessOutput(unittest.TestCase):

    # Note that 'arg_param' is automatically given to the test case, even though it was not
    # part of the test context.
    def test_process_outputs_expected_value(self, proc_output, arg_param):
        proc_output.assertWaitFor('--' + arg_param, timeout=10, stream='stdout')



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading


class ReadyAggregator:
    """Calls a ready_fn parent function on the nth call to a child function."""

    def __init__(self, ready_fn, num_to_aggregate):
        """
        Create a ReadyAggregator.

        :param callable ready_fn: The function to call after n calls to ReadyAggregator.ready_fn
        :param int num_to_aggregate: Number of calls to ReadyAggregator.ready_fn necessary for
        the parent ready_fn to be called
        """
        self._parent_ready_fn = ready_fn
        self._count_to_activate = num_to_aggregate

        self._lock = threading.Lock()

    def ready_fn(self):

        # We don't want to call the parent ready function while holding a lock
        # in case it does something that will try to acquire the same lock.  Instead,
        # we'll make a local copy of the count and evaluate it outside of the critical
        # section
        local_count = 0
        with self._lock:
            self._count_to_activate -= 1
            local_count = self._count_to_activate

        if local_count == 0:
            self._parent_ready_fn()



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

from launch.actions import EmitEvent
from launch.actions import ExecuteProcess
from launch.actions import RegisterEventHandler
from launch.event_handlers import OnExecutionComplete
from launch.event_handlers import OnProcessExit
from launch.event_handlers import OnProcessIO
from launch.event_handlers import OnShutdown
from launch.events import Shutdown

from . import output


class LaunchTestService():

    def __init__(self):
        self.__tests = OrderedDict()
        self.__processes_rc = OrderedDict()

    def _arm(
        self,
        test_name
    ):
        """Prepare for test execution."""
        assert test_name not in self.__tests
        self.__tests[test_name] = 'armed'

    def _finish(
        self,
        test_name
    ):
        """Mark test as finished and shutdown if all other tests have finished too."""
        assert test_name in self.__tests
        self.__tests[test_name] = 'finished'
        if all(status != 'armed' for status in self.__tests.values()):
            return [EmitEvent(event=Shutdown(reason='all tests finished'))]

    def _drop(
        self,
        test_name
    ):
        """Mark test as dropped."""
        assert test_name in self.__tests
        self.__tests[test_name] = 'dropped'

    def _fail(
        self,
        test_name,
        reason
    ):
        """Mark test as a failure and shutdown, dropping the tests that are still in progress."""
        assert test_name in self.__tests
        self.__tests[test_name] = 'failed'
        for test_name in self.__tests:
            if self.__tests[test_name] == 'armed':
                self.__tests[test_name] = 'dropped'
        return [EmitEvent(event=Shutdown(reason=reason))]

    def _succeed(
        self,
        test_name,
        side_effect=None
    ):
        """Mark test as a success and shutdown if all other tests have finished too."""
        assert test_name in self.__tests
        self.__tests[test_name] = 'succeeded'
        if all(status != 'armed' for status in self.__tests.values()):
            return [EmitEvent(event=Shutdown(reason='all tests finished'))]
        if side_effect == 'shutdown':
            return [EmitEvent(event=Shutdown(reason='shutdown after test'))]
        return []

    def add_fixture_action(
        self,
        launch_description,
        action,
        exit_allowed=[0],
    ):
        """
        Add action used as testing fixture.

        If a process action and it exits, a shutdown event is emitted.
        """
        launch_description.add_action(action)
        if isinstance(action, ExecuteProcess):
            def on_fixture_process_exit(event, context):
                process_name = event.action.process_details['name']
                allowed_to_exit = exit_allowed
                if isinstance(exit_allowed, list):
                    allowed_to_exit = event.returncode in exit_allowed
                if not context.is_shutdown and not allowed_to_exit:
                    rc = event.returncode if event.returncode else 1
                    self.__processes_rc[process_name] = rc
                    return EmitEvent(event=Shutdown(
                        reason='{} fixture process died!'.format(process_name)
                    ))
            launch_description.add_action(
                RegisterEventHandler(OnProcessExit(
                    target_action=action, on_exit=on_fixture_process_exit
                ))
            )
        return action

    def add_test_action(
        self,
        launch_description,
        action
    ):
        """
        Add action used for testing.

        If either all test actions have completed or a process action has
        exited with a non-zero return code, a shutdown event is emitted.
        """
        test_name = 'test_{}'.format(id(action))
        if isinstance(action, ExecuteProcess):
            def on_test_process_exit(event, context):
                if event.returncode != 0:
                    process_name = event.action.process_details['name']
                    self.__processes_rc[process_name] = event.returncode
                    return self._fail(
                        test_name, reason='{} test failed!'.format(
                            process_name
                        )
                    )
                return self._succeed(test_name)

            launch_description.add_action(
                RegisterEventHandler(OnProcessExit(
                    target_action=action, on_exit=on_test_process_exit
                ))
            )
        else:
            def on_test_completion(event, context):
                future = event.action.get_asyncio_future()
                if future is not None:
                    if future.cancelled():
                        return self._drop(test_name)
                    exc = future.exception()
                    if exc is not None:
                        return self._fail(test_name, str(exc))
                return self._succeed(test_name)

            launch_description.add_action(
                RegisterEventHandler(OnExecutionComplete(
                    target_action=action, on_completion=on_test_completion
                ))
            )
        launch_description.add_action(action)
        self._arm(test_name)
        return action

    def add_output_test(
        self,
        launch_description,
        action,
        output_test,
        test_suffix='output',
        output_filter=None,
        side_effect=None,
    ):
        """
        Test an action process' output against a given test.

        :param launch_description: test launch description that owns the given action.
        :param action: launch action to test whose output is to be tested.
        :param output_test: test tuple as returned by launch_testing.output.create_* functions.
        :param test_suffix: an optional test suffix to disambiguate multiple test instances,
        defaults to 'output'.
        :param output_filter: an optional function to filter out i.e. ignore output lines for
        the test.
        :param side_effect: an optional side effect of a passing test, currently only 'shutdown'
        is supported.
        """
        assert isinstance(action, ExecuteProcess)
        test_name = 'test_{}_{}'.format(id(action), test_suffix)
        out, collate_output, match_output, match_patterns = output_test
        if not output_filter:
            output_filter = (lambda x: x)
        assert any(match_patterns)

        def on_process_exit(event, context):
            nonlocal match_patterns
            if any(match_patterns):
                # Finish test instead of failing to prevent process exit
                # and process output event handlers from racing.
                return self._finish(test_name)

        launch_description.add_action(
            RegisterEventHandler(OnProcessExit(
                target_action=action, on_exit=on_process_exit
            ))
        )

        def on_shutdown(event, context):
            nonlocal match_patterns
            if any(match_patterns):
                process_name = action.process_details['name']
                reason = 'not all {} output matched!'.format(process_name)
                self._fail(test_name, reason)
            self._succeed(test_name, side_effect)

        launch_description.add_action(
            RegisterEventHandler(OnShutdown(on_shutdown=on_shutdown))
        )

        def on_process_stdout(event):
            nonlocal out
            nonlocal match_patterns

            out = collate_output(out, output_filter(event.text))
            match_patterns = [
                pattern for pattern in match_patterns
                if not match_output(out, pattern)
            ]
            if not any(match_patterns):
                return self._succeed(test_name, side_effect)
            return None

        launch_description.add_action(
            RegisterEventHandler(OnProcessIO(
                target_action=action, on_stdout=on_process_stdout
            ))
        )
        self._arm(test_name)

        return action

    def run(
        self,
        launch_service,
        *args,
        **kwargs
    ):
        """
        Invoke the `run` method of the launch service.

        :returns: If the return value of the parent method is zero but any of
          the test processes exited with a non-zero return code the return of
          the first failed test process is returned.
        """
        rc = launch_service.run(*args, **kwargs)
        if rc == 0:
            default_rc = 1 if 'failed' in self.__tests.values() else 0
            rc = next((rc for rc in self.__processes_rc.values() if rc != 0), default_rc)
        return rc


__all__ = [
    'actions',
    'LaunchTestService',
    'output',
]



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A module providing process info capturing classes.

To prevent pytest from rewriting this module assertions, please PYTEST_DONT_REWRITE.
See https://docs.pytest.org/en/latest/assert.html#disabling-assert-rewriting for
further reference.
"""


import threading
from launch.actions import ExecuteProcess  # noqa
from launch.events.process import ProcessExited

from .util import NoMatchingProcessException
from .util import resolveProcesses


class ProcInfoHandler:
    """Captures exit codes from processes when they terminate."""

    def __init__(self):
        self._proc_info = {}

    def append(self, process_info):
        self._proc_info[process_info.action] = process_info

    def __iter__(self):
        return self._proc_info.values().__iter__()

    def processes(self):
        """Get the ExecuteProcess launch actions of all recorded processes."""
        return self._proc_info.keys()

    def process_names(self):
        """Get the name of all recorded processes."""
        return map(
            lambda x: x.process_details['name'],
            self._proc_info.keys()
        )

    def __getitem__(self, key):
        """
        Get the ProcessExited event for the specified process.

        :param key: Either a string, or a launch.actions.ExecuteProcess object
        :returns launch.events.process.ProcessExited:
        """
        if isinstance(key, str):
            # Look up by process name
            for (launch_action, value) in self._proc_info.items():
                if key in launch_action.process_details['name']:
                    return value
            else:
                raise KeyError(key)
        else:
            return self._proc_info[key]


class ActiveProcInfoHandler:
    """Allows tests to wait on process termination before proceeding."""

    def __init__(self):
        self._sync_lock = threading.Condition()
        # Deliberately not calling the 'super' constructor here.  We're building this class
        # by composition so we can still give out the unsynchronized version
        self._proc_info_handler = ProcInfoHandler()

    @property
    def proc_event(self):
        return self._sync_lock

    def append(self, process_info):
        with self._sync_lock:
            self._proc_info_handler.append(process_info)
            self._sync_lock.notify()

    def __iter__(self):
        with self._sync_lock:
            return self._proc_info_handler.__iter__()

    def processes(self):
        """
        Get the ExecuteProcess launch actions of all recorded processes.

        :returns [launch.actions.ExecuteProcess]:
        """
        with self._sync_lock:
            return list(self._proc_info_handler.processes())

    def process_names(self):
        """
        Get the name of all recorded processes.

        :returns [string]:
        """
        with self._sync_lock:
            return list(self._proc_info_handler.process_names())

    def __getitem__(self, key):
        with self._sync_lock:
            return self._proc_info_handler[key]

    def assertWaitForShutdown(self,
                              process,
                              cmd_args=None,
                              *,
                              timeout=10):

        success = False

        def proc_is_shutdown():
            try:
                resolved_process = resolveProcesses(
                    info_obj=self._proc_info_handler,
                    process=process,
                    cmd_args=cmd_args,
                    strict_proc_matching=True
                )[0]
                process_event = self._proc_info_handler[resolved_process]
                return isinstance(process_event, ProcessExited)
            except NoMatchingProcessException:
                return False

        with self._sync_lock:
            success = self._sync_lock.wait_for(
                proc_is_shutdown,
                timeout=timeout
            )

        assert success, "Timed out waiting for process '{}' to finish".format(process)

    def assertWaitForStartup(self,
                             process,
                             cmd_args=None,
                             *,
                             timeout=10):
        success = False

        def proc_is_started():
            try:
                resolveProcesses(
                    info_obj=self._proc_info_handler,
                    process=process,
                    cmd_args=cmd_args,
                    strict_proc_matching=True
                )
                return True
            except NoMatchingProcessException:
                return False

        with self._sync_lock:
            success = self._sync_lock.wait_for(
                proc_is_started,
                timeout=timeout
            )

        assert success, "Timed out waiting for process '{}' to start".format(process)



#!/usr/bin/python3

# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import time


# This process pretends to do some simple setup, then pretends to do some simple work,
# then shuts itself down automatically
if __name__ == '__main__':

    print('Starting Up')
    time.sleep(1.0)
    print('Ready')

    if sys.argv[1:]:
        print('Called with arguments {}'.format(sys.argv[1:]))

    if '--exception' in sys.argv[1:]:
        raise Exception('Process had a pretend error')

    try:
        print('Emulating Work')
        time.sleep(1.0)
        print('Done')
    except KeyboardInterrupt:
        pass

    print('Shutting Down')

    sys.exit(0)



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def post_shutdown_test():
    """Decorate tests that are meant to run after process shutdown."""
    def decorator(test_item):
        if not isinstance(test_item, type):
            raise TypeError('postcondition_test should decorate test classes')
        test_item.__post_shutdown_test__ = True
        return test_item

    return decorator



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import launch.actions

NO_CMD_ARGS = object()


class NoMatchingProcessException(Exception):
    pass


class _FakeContextException(Exception):
    pass


class _fake_context:

    def __getattr__(self, attr):
        # If anything actually tries to use our fake context, we will raise an exception
        # that we can recognize.
        raise _FakeContextException()


def _proc_to_name_and_args(proc):
    # proc is a launch.actions.ExecuteProcess

    if proc.process_details:
        # process_details are added by the launch system when the process is started.
        # Before the process starts, this will be None
        return '{} {}'.format(
            proc.process_details['name'],
            ' '.join(proc.process_details['cmd'][1:])
        )
    else:
        # Process has not been launched yet, so we don't yet know the process name.
        # In the special case where the process's name is all TextSubstitutions, we
        # can resolve the name.  If there are any other substitutions, we can't give
        # additional info until after the process is launched
        def perform_subs(subs):
            return ''.join([sub.perform(_fake_context()) for sub in subs])
        try:
            cmd = []
            if (isinstance(proc, launch.actions.ExecuteProcess)):
                cmd = [perform_subs(sub) for sub in proc.cmd]
            elif (isinstance(proc, launch.actions.ExecuteLocal)):
                cmd = [perform_subs(sub) for sub in proc.process_description.cmd]
            return ' '.join(cmd)
        except _FakeContextException:
            return 'Unknown - Process not launched yet'


def _str_name_to_process(info_obj, proc_name, cmd_args):

    def name_match_fn(proc):
        return proc_name in proc.process_details['name']

    def cmd_match_fn(proc):
        if cmd_args is None:
            return True
        elif cmd_args is NO_CMD_ARGS:
            return len(proc.process_details['cmd']) == 1
        else:
            return all(arg in proc.process_details['cmd'][1:] for arg in cmd_args)

    matches = [proc for proc in info_obj.processes()
               if name_match_fn(proc) and cmd_match_fn(proc)]

    return matches


def resolveProcesses(info_obj, *, process=None, cmd_args=None, strict_proc_matching=True):
    """
    Resolve a process name and cmd arguments to one or more launch.actions.ExecuteProcess.

    :param info_obj: a ProcInfoHandler or an IoHandler that contains processes that could match

    :param process:  One or more processes to match.  Pass None to match all processes
    :type process: A launch.actions.ExecuteProcess object to match a specific process, or a string
    to search by process name

    :param cmd_args: Optional.  If the process param is a string, the cmd_args will be used to
    disambiguate processes with the same name.  cmd_args=None will match all command arguments.
    cmd_args=launch_testing.asserts.NO_CMD_ARGS will match a process without command-line
    arguments

    :param strict_proc_matching:  Optional.  If the process param is a string and matches multiple
    processes, strict_proc_matching=True will raise an error

    :returns: A list of one or more matching processes taken from the info_obj.  If no processes
    in info_obj match, a NoMatchingProcessException will be raised.
    """
    if process is None:
        # We want to search all processes
        all_procs = info_obj.processes()
        if len(all_procs) == 0:
            raise NoMatchingProcessException('No data recorded for any process')
        return all_procs

    if (isinstance(process, launch.actions.ExecuteProcess) or
       isinstance(process, launch.actions.ExecuteLocal)):
        # We want to search a specific process
        if process in info_obj.processes():
            return [process]
        else:
            raise NoMatchingProcessException(
                'No data recorded for proc {}'.format(_proc_to_name_and_args(process))
            )

    elif isinstance(process, str):
        # We want to search one (or more) processes that match a particular string.  The 'or more'
        # part is controlled by the strict_proc_matching argument

        # Old versions of this let you pass a string for cmd_args, but there was no way to search
        # for multiple arguments when you did that.  For backwards compatibility, we'll wrap string
        # cmd_args in a list
        if isinstance(cmd_args, str):
            cmd_args = [cmd_args]

        matches = _str_name_to_process(info_obj, process, cmd_args)
        if len(matches) == 0:
            names = ', '.join(sorted(_proc_to_name_and_args(p) for p in info_obj.processes()))

            raise NoMatchingProcessException(
                "Did not find any processes matching name '{}' and args '{}'. Procs: {}".format(
                    process,
                    cmd_args,
                    names
                )
            )

        if strict_proc_matching and len(matches) > 1:
            names = ', '.join(sorted(_proc_to_name_and_args(p) for p in info_obj.processes()))
            raise Exception(
                "Found multiple processes matching name '{}' and cmd_args '{}'. Procs: {}".format(
                    process,
                    cmd_args,
                    names
                )
            )
        return list(matches)

    else:
        # Invalid argument passed for 'process'
        raise TypeError(
            "proc argument must be 'ExecuteProcess' or 'str' not {}".format(type(process))
        )



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import unittest

import ament_index_python

import launch
import launch.actions

import launch_testing
import launch_testing.actions
from launch_testing.asserts import assertSequentialStdout

import pytest


@pytest.mark.launch_test
def generate_test_description():
    TEST_PROC_PATH = os.path.join(
        ament_index_python.get_package_prefix('launch_testing'),
        'lib/launch_testing',
        'good_proc.py'
    )

    # This is necessary to get unbuffered output from the process under test
    proc_env = os.environ.copy()
    proc_env['PYTHONUNBUFFERED'] = '1'

    dut_process = launch.actions.ExecuteProcess(
        cmd=[sys.executable, TEST_PROC_PATH],
        env=proc_env, output='screen'
    )

    return launch.LaunchDescription([
        dut_process,

        # Start tests right away - no need to wait for anything
        launch_testing.actions.ReadyToTest(),
    ]), {'dut_process': dut_process}


# These tests will run concurrently with the dut process.  After all these tests are done,
# the launch system will shut down the processes that it started up
class TestGoodProcess(unittest.TestCase):

    def test_count_to_four(self, proc_output):
        # This will match stdout from any process.  In this example there is only one process
        # running
        proc_output.assertWaitFor('Loop 1', timeout=10, stream='stdout')
        proc_output.assertWaitFor('Loop 2', timeout=10, stream='stdout')
        proc_output.assertWaitFor('Loop 3', timeout=10, stream='stdout')
        proc_output.assertWaitFor('Loop 4', timeout=10, stream='stdout')


@launch_testing.post_shutdown_test()
class TestProcessOutput(unittest.TestCase):

    def test_exit_code(self, proc_info):
        # Check that all processes in the launch (in this case, there's just one) exit
        # with code 0
        launch_testing.asserts.assertExitCodes(proc_info)

    def test_full_output(self, proc_output, dut_process):
        # Using the SequentialStdout context manager asserts that the following stdout
        # happened in the same order that it's checked
        with assertSequentialStdout(proc_output, dut_process) as cm:
            cm.assertInStdout('Starting Up')
            for n in range(4):
                cm.assertInStdout('Loop {}'.format(n))
            if os.name != 'nt':
                # On Windows, process termination is always forced
                # and thus the last print in good_proc never makes it.
                cm.assertInStdout('Shutting Down')

    def test_out_of_order(self, proc_output, dut_process):
        # This demonstrates that we notice out-of-order IO
        with self.assertRaisesRegex(AssertionError, "'Loop 2' not found"):
            with assertSequentialStdout(proc_output, dut_process) as cm:
                cm.assertInStdout('Loop 1')
                cm.assertInStdout('Loop 3')
                cm.assertInStdout('Loop 2')  # This should raise



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import unittest

import ament_index_python

import launch
import launch.actions

import launch_testing
import launch_testing.actions
from launch_testing.asserts import assertSequentialStdout

import pytest


def get_test_process_action():
    TEST_PROC_PATH = os.path.join(
        ament_index_python.get_package_prefix('launch_testing'),
        'lib/launch_testing',
        'good_proc.py'
    )
    return launch.actions.ExecuteProcess(
        cmd=[sys.executable, TEST_PROC_PATH],
        name='good_proc.py',
        # This is necessary to get unbuffered output from the process under test
        additional_env={'PYTHONUNBUFFERED': '1'},
    )


# This launch description shows the prefered way to let the tests access launch actions.  By
# adding them to the test context, it's not necessary to scope them at the module level like in
# the good_proc.test.py example
@pytest.mark.launch_test
def generate_test_description():
    dut_process = get_test_process_action()

    ld = launch.LaunchDescription([
        dut_process,

        # Start tests right away - no need to wait for anything
        launch_testing.actions.ReadyToTest(),
    ])

    # Items in this dictionary will be added to the test cases as an attribute based on
    # dictionary key
    test_context = {
        'dut': dut_process,
        'int_val': 10
    }

    return ld, test_context


class TestProcOutput(unittest.TestCase):

    def test_process_output(self, launch_service, proc_info, proc_output, dut):
        # We can use the 'dut' argument here because it's part of the test context
        # returned by `generate_test_description`  It's not necessary for every
        # test to use every piece of the context
        proc_output.assertWaitFor('Loop 1', process=dut, timeout=10, stream='stdout')


@launch_testing.post_shutdown_test()
class TestProcessOutput(unittest.TestCase):

    def test_full_output(self, proc_output, dut):
        # Same as the test_process_output test. launch_testing binds the value of
        # 'dut' from the test_context to the test before it runs
        with assertSequentialStdout(proc_output, process=dut) as cm:
            cm.assertInStdout('Starting Up')
            if os.name != 'nt':
                # On Windows, process termination is always forced
                # and thus the last print in good proc never makes it.
                cm.assertInStdout('Shutting Down')

    def test_int_val(self, int_val):
        # Arguments don't have to be part of the LaunchDescription.  Any object can
        # be passed in
        self.assertEqual(int_val, 10)

    def test_all_context_objects(self, int_val, dut):
        # Multiple arguments are supported
        self.assertEqual(int_val, 10)
        self.assertIn('good_proc.py', dut.process_details['name'])

    def test_all_context_objects_different_order(self, dut, int_val):
        # Put the arguments in a different order from the above test
        self.assertEqual(int_val, 10)
        self.assertIn('good_proc.py', dut.process_details['name'])



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import launch
import launch.actions
import launch_testing.actions
import launch_testing.markers
import pytest


# This function specifies the processes to be run for our test
@pytest.mark.launch_test
@launch_testing.markers.keep_alive
def generate_test_description():
    """Launch a simple process to print 'hello_world'."""
    return launch.LaunchDescription([
        # Launch a process to test
        launch.actions.ExecuteProcess(
            cmd=['echo', 'hello_world'],
            shell=True
        ),
        # Tell launch to start the test
        launch_testing.actions.ReadyToTest()
    ])


# This is our test fixture. Each method is a test case.
# These run alongside the processes specified in generate_test_description()
class TestHelloWorldProcess(unittest.TestCase):

    def test_read_stdout(self, proc_output):
        """Check if 'hello_world' was found in the stdout."""
        # 'proc_output' is an object added automatically by the launch_testing framework.
        # It captures the outputs of the processes launched in generate_test_description()
        # Refer to the documentation for further details.
        proc_output.assertWaitFor('hello_world', timeout=10, stream='stdout')


# These tests are run after the processes in generate_test_description() have shutdown.
@launch_testing.post_shutdown_test()
class TestHelloWorldShutdown(unittest.TestCase):

    def test_exit_codes(self, proc_info):
        """Check if the processes exited normally."""
        launch_testing.asserts.assertExitCodes(proc_info)



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import importlib.util
import logging
import os
import sys

from .junitxml import unittestResultsToXml
from .loader import LoadTestsFromPythonModule
from .print_arguments import print_arguments_of_launch_description
from .test_runner import LaunchTestRunner

_logger_ = logging.getLogger(__name__)


def _load_python_file_as_module(test_module_name, python_file_path):
    """Load a given Python launch file (by path) as a Python module."""
    spec = importlib.util.spec_from_file_location(test_module_name, python_file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def add_arguments(parser):
    """Add arguments to the CLI parser."""
    parser.add_argument('launch_test_file', help='Path to the launch test.')
    parser.add_argument(
        '--package-name', action='store', default=None,
        help='Name of the package the test is in. Useful to aggregate xUnit reports.'
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true', default=False, help='Run with verbose output'
    )
    parser.add_argument(
        '-s', '--show-args', '--show-arguments', action='store_true', default=False,
        help='Show arguments that may be given to the launch test.'
    )
    parser.add_argument(
        'launch_arguments', nargs='*',
        help="Arguments in '<name>:=<value>' format (for duplicates, last one wins)."
    )
    parser.add_argument(
        '--junit-xml', action='store', dest='xmlpath', default=None,
        help='Do write xUnit reports to specified path.'
    )


def parse_arguments():
    parser = argparse.ArgumentParser(
        description='Launch integration testing tool.'
    )
    add_arguments(parser)
    return parser, parser.parse_args()


def run(parser, args, test_runner_cls=LaunchTestRunner):
    # Load the test file as a module and make sure it has the required
    # components to run it as a launch test
    _logger_.debug("Loading tests from file '{}'".format(args.launch_test_file))
    if not os.path.isfile(args.launch_test_file):
        # Note to future reader: parser.error also exits as a side effect
        parser.error("Test file '{}' does not exist".format(args.launch_test_file))

    args.launch_test_file = os.path.abspath(args.launch_test_file)
    launch_test_file_basename = os.path.splitext(os.path.basename(args.launch_test_file))[0]
    if not args.package_name:
        args.package_name = launch_test_file_basename
    test_module = _load_python_file_as_module(args.package_name, args.launch_test_file)

    # This is a list of TestRun objects.  Each run corresponds to one launch.  There may be
    # multiple runs if the launch is parametrized
    test_runs = LoadTestsFromPythonModule(
        test_module, name='{}.{}.launch_tests'.format(
            args.package_name, launch_test_file_basename
        )
    )

    # The runner handles sequcing the launches
    runner = test_runner_cls(
        test_runs=test_runs,
        launch_file_arguments=args.launch_arguments,
        debug=args.verbose
    )

    _logger_.debug('Validating test configuration')

    runner.validate()

    if args.show_args:
        # TODO pete: Handle the case where different launch descriptions take different args?
        print_arguments_of_launch_description(
            launch_description=test_runs[0].get_launch_description()
        )
        return

    _logger_.debug('Running integration test')

    results = runner.run()

    _logger_.debug('Done running integration test')

    if args.xmlpath:
        xml_report = unittestResultsToXml(
            test_results=results, name='{}.{}'.format(
                args.package_name, launch_test_file_basename
            )
        )
        xml_report.write(args.xmlpath, encoding='utf-8', xml_declaration=True)

    # There will be one result for every test run (see above where we load the tests)
    if not all(result.wasSuccessful() for result in results.values()):
        return 1
    return 0


def main():
    logging.basicConfig()

    parser, args = parse_arguments()

    if args.verbose:
        _logger_.setLevel(logging.DEBUG)
        _logger_.debug('Running with verbose output')

    try:
        sys.exit(run(parser, args))
    except Exception as e:
        parser.error(e)
        sys.exit(1)


if __name__ == '__main__':
    main()



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import xml.etree.ElementTree as ET


def unittestResultsToXml(*, name='launch_test', test_results={}):
    """
    Serialize multiple unittest.TestResult objects into an XML document.

    A testSuites element will be the root element of the document.
    """
    # The test_suites element is the top level of the XML result.
    # launch_test results contain two test suites - one from tests that ran while processes
    # were active, and one from tests that ran after processes were shut down
    test_suites = ET.Element('testsuites')
    test_suites.set('name', name)

    # To get tests, failures, and errors, we just want to iterate the results once
    tests = 0
    failures = 0
    errors = 0
    time = 0

    for result in test_results.values():
        tests += result.testsRun
        failures += len(result.failures)
        errors += len(result.errors)
        time += sum(result.testTimes.values())

    test_suites.set('tests', str(tests))
    test_suites.set('failures', str(failures))
    test_suites.set('errors', str(errors))
    test_suites.set('time', str(round(time, 3)))

    for (name, test_result) in test_results.items():
        test_suites.append(unittestResultToXml(str(name), test_result))

    return ET.ElementTree(test_suites)


def unittestResultToXml(name, test_result):
    """
    Serialize a single unittest.TestResult to an XML element.

    <testsuite name="str" tests="#" failures="#" errors="#">
      . . .
    </testsuite>
    """
    test_suite = ET.Element('testsuite')
    test_suite.set('name', name)
    test_suite.set('tests', str(test_result.testsRun))
    test_suite.set('failures', str(len(test_result.failures)))
    test_suite.set('errors', str(len(test_result.errors)))
    test_suite.set('skipped', str(len(test_result.skipped)))
    test_suite.set('time', str(round(sum(test_result.testTimes.values()), 3)))

    for case in test_result.testCases:
        test_suite.append(unittestCaseToXml(test_result, case))

    return test_suite


def unittestCaseToXml(test_result, test_case):
    """
    Serialize a unittest.TestCase into an XML element.

    <testcase name="test_1" time="1.05"/>

    Note - an ordinary unittest.TestResult does not record time information.  The TestResult
    class needs to be a launch_testing.TestResult class
    """
    case_xml = ET.Element('testcase')
    full_methodname, _, qualifiers = test_case.id().partition(' ')
    full_classname, _, methodname = full_methodname.rpartition('.')
    case_xml.set('classname', full_classname)
    case_xml.set('name', (methodname + ' ' + qualifiers).strip())
    case_xml.set('time', str(round(test_result.testTimes[test_case], 3)))

    for failure in test_result.failures:
        # We're enumerating a list of (test_case, failure_string) tuples here
        if failure[0] == test_case:
            failure_xml = ET.Element('failure')
            failure_xml.set('message', failure[1])
            case_xml.append(failure_xml)

    for error in test_result.errors:
        # Same as above.  (test_case, error_string) tuples
        if error[0] == test_case:
            error_xml = ET.Element('error')
            error_xml.set('message', error[1])
            case_xml.append(error_xml)

    for skip in test_result.skipped:
        if skip[0] == test_case:
            skip_xml = ET.Element('skipped')
            skip_xml.set('message', skip[1])
            case_xml.append(skip_xml)

    return case_xml



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from typing import Optional
from typing import Text
from typing import Tuple

from launch.actions import ExecuteProcess
from launch.event_handlers import OnProcessIO
from launch.some_actions_type import SomeActionsType


class StdoutReadyListener(OnProcessIO):
    """
    Part of a LaunchDescription that can wait for processes to signal ready with stdout.

    Some processes signal that they're ready by printing a message to stdout.  This listener
    can be added to a launch description to wait for a particular process to output a particular
    bit of text
    """

    def __init__(
        self,
        *,
        target_action: Optional[ExecuteProcess] = None,
        ready_txt: Text,
        actions: [SomeActionsType]
    ):
        self.__ready_txt = ready_txt
        self.__actions = actions

        super().__init__(
            target_action=target_action,
            on_stdout=self.__on_stdout
        )

    def __on_stdout(self, process_io):
        if self.__ready_txt in process_io.text.decode():
            return self.__actions

    def describe(self) -> Tuple[Text, List[SomeActionsType]]:
        """Return the description list with 0 as a string, and then LaunchDescriptionEntity's."""
        description = super().describe()[0]
        return (
            description,
            self.__actions
        )
